# Read the input FIFO data and format it as a line of text
# copied to version 1 in the backup directory
# assumes the A32 pointer is set to the base address
# Now to read and understand the data coming through the main fifo 

global EGreport
# log flag set to 0 stops the writing VME access reports to the log window
# EGreport set to -1 probably stops any errors causing popups ??
set LogFlag 0
set EGreport -1
set log_data 0			;# set to inhibit logging the data from the main fifo

# input FIFO details
set time_ls	0x500
set time_ms	0x504
set ip_count	0x508
set grpA	0x510
set grpB	0x520
set ip_type	0x530

#Middle FIFO details
set fdata	0x0010
set track	0x0014
set fflags	0x0018 

set fileId 0
set file_ok 0
# type table 
set ip_typetable(0)	GroupA
set ip_typetable(1)	GroupB
set ip_typetable(2)	SYNC
set ip_typetable(3)	Pause
set ip_typetable(4)	Resume
set ip_typetable(5)	GroupA_Gated
set ip_typetable(6)	GroupB_Gated

# error messages table
set errortable(0)	"0:this is not an error"
set errortable(1)	"1:The Id is not recognised, could be time ??"
set errortable(2)	"2:The Time does not have top four bits clear"
set errortable(3)	"3:The Time is not greater than or equal to previous in this Data item"
set errortable(4)	"4:The Time is not greater than previous in this Info item"
set errortable(5)	"5:Not an expected Id"
set errortable(6)	"6:Shouldn't reach this point "
set errortable(7)	"7:this is not an error"


#variables for keeping track of time ;-)
set tip_old_m 	0
set tip_old_l 	0 
set tip_tim_ok 	0 
set ip_idx 	0
set sync_top 	0
set last_tim_l 	0 

# variables for keeping track of main data flow
set track_old 0
set first_one 0 
set main_idx 0
set last_id 3 
set last_mf_id 0 
set last_mf_time 0
 
proc file_setup {} {
global fileId file_ok
	if [catch { open /MIDAS/vme/tcl/diag-scripts/pr_fifo.txt {RDWR CREAT} } fileId ] {
		log "Cannot open file pr_fifo.txt: $fileId"
		set file_ok 0
	} else {
		set file_ok 1
	}
}
	
proc time48_check {tim_bot tim_top } {
global tip_old_l tip_old_m tip_tim_ok ip_idx
	if { $tim_bot > $tip_old_l  && $tim_top == $tip_old_m  || $tim_top >= $tip_old_m} {
		set tip_old_l $tim_bot
		set tip_old_m $tim_top
		set tip_tim_ok 1
	} else {
		log "Index: [format "%d" $ip_idx ] => time error"
		set tip_tim_ok 0 
	}	
}
	
# read input fifo and format for one line
proc inp_read {} {
global time_ms time_ls grpA grpB ip_type ip_count fileId file_ok ip_idx
	set time_top [scan [ rdl $time_ms ] {%x} ]
	set time_bot [scan [rdl $time_ls ]  {%x} ]
	set A_patt [scan [ rdl $grpA] {%x} ]
	set B_patt [scan [ rdl $grpB] {%x} ]
	set typ_val [scan [rdl $ip_type] {%x} ]
	time48_check $time_bot $time_top
	if { $file_ok == 1 } {
		puts $fileId "Input: T:[format "%04X%07X" $time_top $time_bot ], A:[format "%04X" $A_patt], B: [format "%04X" $B_patt], Type:[format "%02X" $typ_val]"
	} else {
	log "Input: T:[format "%04X%07X" $time_top $time_bot ], A:[format "%04X" $A_patt], B: [format "%04X" $B_patt], Type:[format "%02X" $typ_val]"
	}
}

# read main fifo and output it with the track
proc main_read {} {
global fdata track fflags file_ok track_old first_one main_idx log_data sync_top last_tim_l errortable
global last_mf_id last_mf_time

	set mf_id [scan [ rdl $fdata] {%x} ] 
	set mf_track [scan [ rdl $track ] {%x} ]
	set rdo_track [expr $mf_track >> 4  & 15 ]
	if {$first_one == 0 } { 
		set track_old $rdo_track
	} else {
		set track_check [ expr $track_old + 2  & 15 ]
		if { $track_check != $rdo_track } { 
			log " Track error .... resuming with current track "
		}
		set track_old $rdo_track
	}	
	set mf_time [scan [ rdl $fdata] {%x} ]
	set main_error [ main_check $mf_id $mf_time ]	;# go and check if this is good data 
	if { $main_error != 0 } { 	;# if it isn't a zero then must stop and look at it ;-(
		log $errortable($main_error) 	;# log a friendly message
		log "Main..Data:[format "%08X : %08X" $mf_id $mf_time ]"
		log "Main..Last:[format "%08X : %08X" $last_mf_id $last_mf_time ]"
		log "Last time :[format "%05X%07X" $sync_top $last_tim_l ]"
		return 1				;# it's all over here 
	}
	if { $log_data == 0 } {
		log "Main.. Track:[format "%d" $rdo_track ], Data:[format "%08X : %08X" $mf_id $mf_time ]"
	}
	set last_mf_id $mf_id
	set last_mf_time $mf_time
	return 0 					;# ok.
}
proc main_check {mc_id mc_tim} {
global	sync_top last_tim_l last_id
	set id [expr $mc_id >> 30 & 3 ]	;# check the id is an id
	if { $id != 3 && $id != 2} {
		return 1
	}
	set tim_check [ expr $mc_tim >> 28 & 15]	;# check the time has zero in the top 4 bits
	if { $tim_check != 0 } {
		return 2
	}
# now we know the data is in the correct order, find out what type it is 
	if { $id == 3 } {			;# then it is an ADC/pattern word, check the time is in order
		if { $mc_tim >= $last_tim_l } {	;# then this is ok, just save the time, and return
			set last_tim_l $mc_tim
			set last_id $id
			return 0 
		} else {			;# error so return a code to say what type 
			return 3	
		}
	} elseif { $id == 2 } {			;# its an information word, check which type, and extract the time top
		set info_type [ expr $mc_id >> 20 & 15 ]
		set info_field [ expr $mc_id & 0x7FFFF ]
		if { $info_type == 2 || $info_type == 3 || $info_type == 4 } {
			if { $info_field == $sync_top && $mc_tim > $last_tim_l } { ;# it's all ok ...
				set last_tim_l $mc_tim
				set sync_top $info_field
				set last_id $id
				return 0 
			} elseif { $info_field == $sync_top && $mc_tim == $last_tim_l && $last_id == 3 } { 
				set last_tim_l $mc_tim				;# it's ok cos last was data ...
				set sync_top $info_field
				set last_id $id
				return 0 
			} elseif { $info_field > $sync_top } { 			;# it's all ok too ...
				set last_tim_l $mc_tim
				set sync_top $info_field
				set last_id $id
				return 0 
			} else {
				return 4 ;# time is all wrong 
			}
		} else { 	;# it's not an expected id ... no pile-up in the pattern register
			return 5 
		}
	} else {
		return 6	;# shouldn't get here as it's filtered earlier 
	}
	return 7 		;# again shouldn't get to this point i don't think 
	
}

proc dump_it {dump_length} {
global ip_count ip_idx tip_old_m tip_old_l sync_top last_tim_l 
	set dot_count 0 
	inp_read
	for { set i 0 } { $i < $dump_length } { incr i } {
		set dot_count [ expr $dot_count + 1 ]
		set dump_error [ main_read ]
		if { $dump_error != 0 } {
			log "Error occured at index:[format "%d" $i] through the main FIFO "
			return 1			;# something went wrong
		}
		if { $i == 1 } { 
			log "Started at Input:[format "%04X%07X" $tip_old_m $tip_old_l ]" 
			log "           Main :[format "%04X%07X" $sync_top $last_tim_l ]" 
		}
		if { $dot_count == 10001 } {
			set dot_count 0 
			log "done : [format "%d" $i ] reads "
		}
		update idletasks
	}
	log "Ended at   Main :[format "%04X%07X" $sync_top $last_tim_l ]" 
}
# file_setup
set log_data 1		;# disable regular printing of the main data 
dump_it 99999999
# close $fileId

